module Language.Java.Paragon.TypeCheck.Monad 
    (
     check, ignore, orElse, maybeM,
     withFold, withFoldMap,

     getThisType, getReturn,
     extendVarEnvList, extendVarEnv, lookupVar, lookupActorName,

--     lookupField,
     lookupFieldT,
     lookupMethod, lookupMethodT,
     lookupConstr, 
     lookupLockArity, 
     lookupExn, registerExn, registerExns,

     extendLockEnv,

     getBranchPC, getBranchPC_, extendBranchPC, 
     addBranchPC, addBranchPCList,

     getActorId, setActorId, 
     newActorId, newActorIdWith, newAliasId,
     freshActorId, aliasActorId,
     scrambleActors,

     getExnPC, throwExn,
     activateExns, deactivateExn,
     getExnState, mergeActiveExnStates, useExnState,

     getCurrentLockState, applyLockMods,
     openLock, closeLock,

     newMetaPolVar,

     constraint, constraint_,
     exnConsistent,

     extendTypeMapP, extendTypeMapT, lookupPkgTypeMap,
     getTypeMap, withTypeMap,

     evalSrcType, evalSrcRefType, evalSrcClsType,
     evalSrcTypeArg, evalSrcNWTypeArg,
     evalReturnType,

     getReadPolicy, getWritePolicy, getLockPolicy,
     getParamPolicy, getReturnPolicy,

     fromSrcType,

     evalPolicy, evalPolicyExp,
     evalLock, evalActor,

     Tc, (|||), getState, mergeWithState, liftBase,

     runTc, setupStartState,

     TcBase, runTcBase, liftIO, withErrCtxt

    ) where

import Language.Java.Paragon.Syntax
import Language.Java.Paragon.Pretty

import Language.Java.Paragon.TypeCheck.Monad.TcBase
import Language.Java.Paragon.TypeCheck.Monad.TcMonad

import Language.Java.Paragon.TypeCheck.TcEnv
import Language.Java.Paragon.TypeCheck.TcState
import Language.Java.Paragon.TypeCheck.Uniq
import Language.Java.Paragon.TypeCheck.Types
import Language.Java.Paragon.TypeCheck.Actors
import Language.Java.Paragon.TypeCheck.Policy
import Language.Java.Paragon.TypeCheck.Locks
import Language.Java.Paragon.TypeCheck.Constraints
import Language.Java.Paragon.TypeCheck.Containment
--import Language.Java.Paragon.TypeCheck.Interpreter

import Control.Monad (liftM, msum)         -- liftM only to instantiate fmap
import Control.Applicative ( (<$>) ) -- ... so we can use <$> instead
import Control.Arrow ( first, second )
import qualified Data.Map as Map
import Data.IORef
import Data.List (union, intersperse)

debug str = liftIO $ putStrLn $ "DEBUG: " ++ str
debugTc = liftBase . debug

--------------------------------------------
--                                        --
--      Monad-independent helpers         --
--                                        --
--------------------------------------------

check :: Monad m => Bool -> String -> m ()
check b err = if b then return () else fail err

ignore :: Monad m => m a -> m ()
ignore = (>> return ())

orElse :: Monad m => m (Maybe a) -> m a -> m a
orElse monma mona = do
  ma <- monma
  case ma of
    Just a -> return a
    Nothing -> mona

infixr 3 `orElse`

maybeM :: Monad m => Maybe a -> (a -> m ()) -> m ()
maybeM ma f = maybe (return ()) f ma

withFold :: Monad m => [m a -> m a] -> m a -> m a
withFold = foldr (.) id

withFoldMap :: Monad m => (a -> m b -> m b) -> [a] -> m b -> m b
withFoldMap f = withFold . map f



--------------------------------------------
--          Working with the env          --
--------------------------------------------

-- this

getThisType :: Tc TcType
getThisType = liftBase getThisT

-- returns

getReturn :: Tc (TcType, TcPolicy)
getReturn = returnI <$> getEnv

{-
registerReturn :: TcType -> TcPolicy -> Tc a -> Tc a
registerReturn ty p = withEnv $ \env -> env { returnI = (ty,p) }
-}
-- lookup entities

--fieldEnvToVars :: Tc a -> Tc a
--fieldEnvToVars = withEnv $ \env -> env { vars = fields (typemap env) }


extendVarEnvList :: [(Ident, VTypeInfo)] -> Tc a -> Tc a
extendVarEnvList vs = withEnv $ \env ->
  let oldVmap = vars env
      newVmap = foldl (\m (i,vti) -> Map.insert i vti m) oldVmap vs
  in env { vars = newVmap }

extendVarEnv :: Ident -> VTypeInfo -> Tc a -> Tc a
--extendVarEnv i = extendVarEnvN i
--extendVarEnvN :: Ident -> VTypeInfo -> Tc a -> Tc a
extendVarEnv i vti = withEnv $ \env ->
  let oldVmap = vars env
      newVmap = Map.insert i vti oldVmap
  in env { vars = newVmap }

lookupActorName :: ActorName -> Tc (TcType, TcPolicy)
lookupActorName (ActorName n) = lookupVar n
lookupActorName (ActorTypeVar i) = lookupVar (Name [i])

lookupVar :: Name -> Tc (TcType, TcPolicy)
lookupVar (Name n@(i:is)) = do
  tm <- liftBase getTypeMap
  (newTm,ty,pol,rest) <- do varMap <- vars <$> getEnv
                            case Map.lookup i varMap of
                              Just (VTI ty p _ _) -> do
                                let tyTm = lookupTypeOfT ty tm
                                return (tyTm, ty, p, is)
                              Nothing -> return (tm,voidT,bottom,n)
  aux newTm ty pol rest

 where aux :: TypeMap -> TcType -> TcPolicy -> [Ident] -> Tc (TcType, TcPolicy)
       aux _  ty pol [] = return (ty, pol)
       aux tm _  pol (i:is) = do
         (newTm,newTy,newPol) <- 
             case Map.lookup i (fields tm) of
               Just (VTI typ pol _ _) -> do
                 baseTm <- liftBase getTypeMap
                 return $ (lookupTypeOfT typ baseTm, typ, pol)
               Nothing -> case Map.lookup i (pkgsAndTypes tm) of
                            Just aTm -> return (aTm,voidT,bottom)
                            Nothing -> fail $ "lookupVar: No such name: " ++ show (i:is)
         aux newTm newTy (pol `join` newPol) is


lookupMethod :: Name -> [TcType] -> Tc (TcPolicy, [TypeParam], MTypeInfo)
lookupMethod (Name n@(i:is)) ts = do
  tm <- liftBase getTypeMap
  (newTm,pol,rest) <- do varMap <- vars <$> getEnv
                         case Map.lookup i varMap of
                           Just (VTI ty p _ _) -> do
                             let tyTm = lookupTypeOfT ty tm
                             return (tyTm, p, is)
                           Nothing -> return (tm,bottom,n)
  aux newTm pol rest ts

 where aux :: TypeMap -> TcPolicy -> [Ident] -> [TcType] -> Tc (TcPolicy, [TypeParam], MTypeInfo)
       aux tm pol [i] ts =
           case Map.lookup (i,ts) (methods tm) of
             Just (tps, mti) -> return (pol, tps, mti)
             Nothing -> fail $ "lookupMethod: No such name: " ++ show i
                          ++ "\nMethods: " ++ show (methods tm)
       aux tm pol (i:is) ts = do
         (newTm,newPol) <- 
             case Map.lookup i (fields tm) of
               Just (VTI typ pol _ _) -> do
                 baseTm <- liftBase getTypeMap
                 return $ (lookupTypeOfT typ baseTm, pol)
               Nothing -> case Map.lookup i (pkgsAndTypes tm) of
                            Just aTm -> return (aTm,bottom)
                            Nothing -> error "lookupTypeOfN: No such name"
         aux newTm (pol `join` newPol) is ts

--  tm <- liftBase getTypeMap
--  case lookupNamedMethod n ts tm of
--    Just tpsmti -> return tpsmti
--    Nothing -> fail $ "Unknown method: " ++ prettyPrint n 
--                       ++ "(" ++ concat (intersperse "," $ map show ts) ++ ")"


getPolicy :: Name -> Tc TcPolicy
getPolicy n = do
  tm <- liftBase getTypeMap
  case lookupNamed policies n tm of
    Nothing -> fail $ "getPolicy: No such policy: " ++ show n
    Just p -> return p


lookupFieldT :: TcType -> Ident -> Tc VTypeInfo
lookupFieldT typ i = do
  check (isClassType typ) $ "Not a class type: " ++ show typ
  tm <- liftBase getTypeMap
  let aTm = lookupTypeOfT typ tm
  case Map.lookup i (fields aTm) of
    Just vti -> return vti
    Nothing -> fail $ "Class " ++ prettyPrint (typeName_ typ)
                 ++ " does not have a field '" ++ prettyPrint i ++ "'"
  
lookupMethodT :: TcType -> Ident -> [TcType] -> Tc ([TypeParam], MTypeInfo)
lookupMethodT typ i pts = do
  check (isClassType typ) $ "Not a class type: " ++ show typ
  tm <- liftBase getTypeMap
  let aTm = lookupTypeOfT typ tm
  case Map.lookup (i,pts) (methods aTm) of
    Just pmti -> return pmti
    Nothing -> fail $ "Class " ++ prettyPrint (typeName_ typ)
                 ++ " does not have a method '" ++ prettyPrint i ++ "'"

lookupConstr :: TcClassType -> [TcType] -> Tc ([TypeParam], CTypeInfo)
lookupConstr ctyp pts = do
  tm <- liftBase getTypeMap
  let typ = clsTypeToType ctyp
      aTm = lookupTypeOfT typ tm
  case Map.lookup pts (constrs aTm) of
    Just pcti -> return pcti
    Nothing -> fail $ "Class " ++ prettyPrint (typeName_ typ)
                 ++ " does not have a constructor with argument types "
                 ++ show pts

lookupLockArity :: Name -> Tc Int
lookupLockArity n = do
  tm <- liftBase getTypeMap
  case lookupNamed lockArities n tm of
    Nothing -> fail $ "Unknown lock: " ++ prettyPrint n
    Just arity -> return arity

lookupExn :: TcType -> Tc (TcPolicy, TcPolicy)
lookupExn tyX = do
  exnMap <- exnsE <$> getEnv
  case Map.lookup tyX exnMap of
    Just ps -> return ps
    Nothing -> fail $ "lookupExn: Unregistered exception type: " ++ show tyX

registerExn :: TcType -> TcPolicy -> TcPolicy -> Tc a -> Tc a
registerExn tyX rX wX = registerExns [(tyX, (rX,wX))]

registerExns :: [(TcType, (TcPolicy, TcPolicy))] -> Tc a -> Tc a
registerExns tysPols =
    withEnv $ \env -> 
        let oldMap = exnsE env
            newMap = foldl (\m (t,ps) -> Map.insert t ps m) oldMap tysPols
        in env { exnsE = newMap }


extendLockEnv :: [TcLock] -> Tc a -> Tc a
extendLockEnv locks = withEnv $ \env -> 
                        env { lockstate = lockstate env `union` locks }

getBranchPC :: Entity -> Tc TcPolicy
getBranchPC e = do env <- getEnv
                   return $ branchPC (Just e) env

getBranchPC_ :: Tc TcPolicy
getBranchPC_ = do env <- getEnv
                  return $ branchPC Nothing env

extendBranchPC :: TcPolicy -> Tc a -> Tc a
extendBranchPC = withEnv . joinBranchPC

addBranchPCList :: [Ident] -> Tc a -> Tc a
addBranchPCList is =
    withEnv $ \env ->
        let (bm, def) = branchPCE env
            newBm = foldl (\m i -> Map.insert (VarEntity (Name [i])) bottom m) bm is
        in env { branchPCE = (newBm, def) }

addBranchPC :: Entity -> TcPolicy -> Tc a -> Tc a
addBranchPC ent pol = 
    withEnv $ \env -> 
        let (bm, def) = branchPCE env
            newBm = Map.insert ent pol bm
        in env { branchPCE = (newBm, def) }

{-- Tying the knot for member policies
setFieldPol :: Ident -> TcPolicy -> Tc ()
setFieldPol i pF = do
  fMap <- (fields . typemap) <$> getEnv
  case Map.lookup (Name [i]) fMap of
    Nothing -> fail "panic: lookup of field failed when tying the knot"
    Just (VTI ty tempP stc fin) -> do
      case tempP of
        TcPolicyVar (TcMetaVar i ref) -> do
          mpol <- lift $ readIORef ref
          case mpol of
            Nothing -> do lift $ writeIORef ref (Just pF)
            Just _  -> fail "panic: field policy already written when tying the knot"
        _ -> fail "panic: field policy already set when tying the knot"
      
setMethodPols :: Ident -> TcPolicy -> TcPolicy -> Tc ()
setMethodPols i retM 
-}

--------------------------------------------
--          Working with states           --
--------------------------------------------
{-
-- Leave actor mappings from fields
startState :: Tc ()
startState = updateState $ \s -> s { lockMods = noMods, exnS = Map.empty }
-}

-- Actor Analysis

getActorId :: Name -> Tc ActorId
getActorId n = do
  actorMap <- actorSt <$> getState
  case Map.lookup n actorMap of
    Nothing -> liftBase $ aliasActorId
    Just ai -> return $ aID ai

setActorId :: Name -> ActorId -> Tc ()
setActorId n aid = updateState setAct
  where setAct :: TcState -> TcState
        setAct s@(TcState { actorSt = actMap }) = s { actorSt = Map.adjust replId n actMap }
        replId (AI _ st) = AI aid st


newActorIdWith :: Ident -> ActorId -> Tc ()
newActorIdWith i aid = do
  updateState (insertAct aid)
      where insertAct :: ActorId -> TcState -> TcState
            insertAct aid s@(TcState { actorSt = actMap }) = 
                s { actorSt = Map.insert (Name [i]) (AI aid Stable) actMap }

newActorId :: Ident -> Tc ActorId
newActorId i = do 
  aid <- liftBase freshActorId
  newActorIdWith i aid
  return aid

newAliasId :: Ident -> Tc ActorId
newAliasId i = do
  aid <- liftBase aliasActorId
  newActorIdWith i aid
  return aid


freshActorId, aliasActorId :: TcBase ActorId
freshActorId = (liftIO . newFresh) =<< getUniqRef
aliasActorId = (liftIO . newAlias) =<< getUniqRef
  

scrambleActors :: Maybe TcType -> Tc ()
scrambleActors mtc = do
  let stb = maybe FullV (FieldV . typeName_) mtc
  state <- getState
  u <- liftBase getUniqRef
  newState <- liftBase . liftIO $ scramble u stb state
  setState newState

-- Exception tracking

getExnPC :: Tc TcPolicy
getExnPC = exnPC <$> getState

throwExn :: ExnType -> TcPolicy -> Tc ()
throwExn et pX = do
  uref <- liftBase getUniqRef
  state <- getState
  let oldXmap = exnS state
      newXmap = Map.fromList [(et, ExnPoint state pX)]
  mergedXmap <- liftBase . liftIO $ mergeExns uref oldXmap newXmap
  setState $ state { exnS = mergedXmap }
  
deactivateExn :: ExnType -> Tc ()
deactivateExn et = 
    updateState $ \s -> 
        let oldEmap = exnS s
            newEmap = Map.delete et oldEmap
        in s { exnS = newEmap }
  

activateExns :: [(ExnType, ExnSig)] -> Tc ()
activateExns exns = do
  uref <- liftBase getUniqRef
  state <- getState
  let (ts, sigs) = unzip exns
  xPoints <- mapM (\sigX -> do
                     s <- liftBase . liftIO $ scramble uref FullV $ -- FullV should be Nothing - TODO
                           state { lockMods = lockMods state ||>> exnMods sigX }
                     return (s, exnWrites sigX)) sigs
  let oldXmap = exnS state
      newXmap = Map.fromList $ zip ts (map (uncurry ExnPoint) xPoints)
  mergedXmap <- liftBase . liftIO $ mergeExns uref oldXmap newXmap
  setState $ state { exnS = mergedXmap }

getExnState :: ExnType -> Tc (Maybe TcState)
getExnState et = do
  eMap <- exnS <$> getState
  return $ epState <$> Map.lookup et eMap

mergeActiveExnStates :: Tc ()
mergeActiveExnStates = do
  exnMap <- exnS <$> getState
  mapM_ (mergeWithState . epState) $ Map.elems exnMap

useExnState :: TcState -> Tc ()
useExnState es = updateState (\s -> s { exnS = exnS es })

-- Lockstate tracking

getCurrentLockState :: Tc [TcLock]
getCurrentLockState = do
  base <- lockstate <$> getEnv
  mods <- lockMods <$> getState
  return $ snd $ ([], base) ||>> mods

applyLockMods :: LockMods -> Tc ()
applyLockMods lms = do
  updateState $ \s -> s { lockMods = lockMods s ||>> lms }

openLock, closeLock :: TcLock -> Tc ()
openLock  l = applyLockMods ([],[l])
closeLock l = applyLockMods ([l],[])

-- Policy inference

newMetaPolVar :: Tc TcPolicy
newMetaPolVar = do
  uniq <- liftBase $ liftIO . getUniq =<< getUniqRef
  ref  <- liftBase $ liftIO (newIORef Nothing)
  return $ TcPolicyVar (TcMetaVar uniq ref)

--newRigidPolVar :: Tc TcPolicy
--newRigidPolVar = do
--  uniq <- lift . getUniq =<< getUniqRef
--  return $ TcRigidVar uniq

--------------------------------------------
--       Working with constraints         --
--------------------------------------------

constraint :: [TcLock] -> TcPolicy -> TcPolicy -> Tc ()
constraint ls p1 p2 = do -- addConstraint $ LRT ls p1 p2
  g <- getGlobalLockProps
  case lrt (g `recmeet` rls) p1 p2 of
    Left b -> do
      --debugTc $ "constraint: p1: " ++ show p1
      --debugTc $ "constraint: p2: " ++ show p2
      check b $ "Cannot solve constraint: " ++ show (LRT (g `recmeet` rls) p1 p2)
    Right c -> addConstraint c
  where rls = locksToRec ls

getGlobalLockProps :: Tc TcPolicyRec
getGlobalLockProps = return (TcPolicyRec [])

locksToRec :: [TcLock] -> TcPolicyRec
locksToRec ls = TcPolicyRec (map (\x -> TcClause x []) (map lockToAtom ls))


constraint_ :: TcPolicy -> TcPolicy -> Tc ()
constraint_ = constraint []

exnConsistent :: TcType -> ExnSig -> Tc ()
exnConsistent exnTy (ExnSig rX wX _) = do
    exnMap <- exnsE <$> getEnv
    --debugTc $ "Using exnMap: " ++ show exnMap
    case Map.lookup exnTy exnMap of
      Nothing -> fail $ "Unchecked exception: " ++ prettyPrint (typeName_ exnTy)
      Just (rE, wE) -> do
        constraint_ wX wE
        constraint_ rE rX

{-
fetchConstraints :: Tc () -> Tc [Constraint]
fetchConstraints (Tc f) = 
    Tc $ \e s u -> do 
      ma <- f e s u
      case ma of
        Left errs -> return $ Left errs
        Right ((),s',cs) -> return $ Right (cs,s',[])

solveConstraints :: [Constraint] -> Tc ()
solveConstraints = lift . solve
-}
--------------------------------------------
--       Working with typemaps            --
--------------------------------------------

extendTypeMapP :: [Ident] -> TypeMap -> TcBase a -> TcBase a
extendTypeMapP [] leafTm = withTypeMap (const leafTm) -- WARNING: overwrites
extendTypeMapP is leafTm = 
    withTypeMap $ \tm ->
        go is leafTm tm
            where go :: [Ident] -> TypeMap -> TypeMap -> TypeMap
                  go [i] leafTm tm =
                      let mTm = packages tm in
                      case Map.lookup i mTm of
                        Nothing -> tm { packages = Map.insert i leafTm mTm }
                        Just _ -> tm -- TODO: This means we were already here
                  go (i:is) leafTm tm =
                      let mTm = packages tm 
                          eTm = case Map.lookup i mTm of
                                    Just innerTm -> innerTm
                                    Nothing -> emptyTM
                          newTm = go is leafTm eTm
                      in tm { packages = Map.insert i newTm mTm }

extendTypeMapT :: Ident -> [TypeParam] -> TypeMap -> TcBase a -> TcBase a
extendTypeMapT i tps leafTm =
    withTypeMap $ \tm -> 
        tm { types = Map.insert i (tps, leafTm) (types tm) }


lookupPkgTypeMap :: [Ident] -> TcBase TypeMap
lookupPkgTypeMap is = lookupTypeOfN (Name is) <$> getTypeMap

-------------------------------------------------------------------
-- Evaluating types

evalReturnType :: Maybe Type -> TcBase TcType
evalReturnType = maybe (return voidT) evalSrcType

evalSrcType :: Type -> TcBase TcType
evalSrcType (PrimType pt) = return $ TcPrimT pt
evalSrcType (RefType rt)  = TcRefT <$> evalSrcRefType rt

evalSrcRefType :: RefType -> TcBase TcRefType
evalSrcRefType (TypeVariable i) = return $ TcTypeVar i
evalSrcRefType (ArrayType t mp) = do
  ty <- evalSrcType t
  mpol <- maybe (return Nothing) (fmap Just . evalPolicy) mp
  return $ TcArrayT ty mpol
evalSrcRefType (ClassRefType ct) = TcClsRefT <$> evalSrcClsType ct

evalSrcClsType :: ClassType -> TcBase TcClassType
evalSrcClsType (ClassType iArgs) = 
    TcClassT <$> mapM (\(i,tas) -> (\ts -> (i, ts)) <$> mapM evalSrcTypeArg tas) iArgs
      
evalSrcTypeArg :: TypeArgument -> TcBase TcTypeArg
evalSrcTypeArg (ActualArg a) = evalSrcNWTypeArg a
evalSrcTypeArg _ = fail "evalSrcTypeArg: Wildcards not yet supported"

evalSrcNWTypeArg :: NonWildTypeArgument -> TcBase TcTypeArg
evalSrcNWTypeArg (ActualType rt) = TcActualType <$> evalSrcRefType rt
evalSrcNWTypeArg (ActualPolicy p) = TcActualPolicy <$> evalPolicy p
evalSrcNWTypeArg (ActualActor n) = TcActualActor <$> evalActorId n
evalSrcNWTypeArg (ActualLockState ls) = TcActualLockState <$> mapM evalLock ls

evalPolicy :: Exp -> TcBase TcPolicy
evalPolicy e = case e of
  ExpName n -> do
    --debug $ "evalPolicy: " ++ show n
    tm <- getTypeMap
    case lookupNamed policies n tm of
      Just p -> return p
      Nothing -> fail $ "evalPolicy: no such policy\n" ++ "TypeMap: " 
                     ++ show (types tm)
  PolicyExp pl -> evalPolicyExp pl
  BinOp p1 Add p2 -> do
    pol1 <- evalPolicy p1
    pol2 <- evalPolicy p2
    return $ pol1 `meet` pol2
  BinOp p1 Mult p2 -> do
    pol1 <- evalPolicy p1
    pol2 <- evalPolicy p2
    return $ pol1 `join` pol2
  Paren p -> evalPolicy p
  _ -> fail "evalPolicy: More here!"

evalPolicyExp :: PolicyExp -> TcBase TcPolicy
evalPolicyExp (PolicyLit cs) = TcPolicy <$> mapM evalClause cs
evalPolicyExp (PolicyOf i)   = return $ TcRigidVar i
evalPolicyExp (PolicyTypeVar i) = return $ TcRigidVar i

evalClause :: Clause Actor -> TcBase (TcClause TcActor)
evalClause (Clause h b) = do
  h' <- evalActor h
  b' <- mapM evalAtom b
  return $ TcClause h' b'

evalActorName :: ActorName -> TcBase ActorId
evalActorName (ActorName n) = evalActorId n
evalActorName (ActorTypeVar i) = return $ ActorTPVar i

evalActor :: Actor -> TcBase TcActor
evalActor (Actor n) = TcActor <$> evalActorName n
evalActor (Var i) = return $ TcVar i

evalActorId :: Name -> TcBase ActorId
evalActorId n = do
  tm <- getTypeMap
  case lookupNamed actors n tm of
    Just aid -> return aid
    Nothing -> fail $ "evalActor: No such actor: " ++ prettyPrint n

evalAtom :: Atom -> TcBase TcAtom
evalAtom (Atom n as) = TcAtom n <$> mapM evalActor as

evalLock :: Lock -> TcBase TcLock
evalLock (Lock n ans) = do
  tm <- getTypeMap
  case lookupNamed lockArities n tm of
    Just arL -> do
      check (length ans == arL) $
                "Lock " ++ prettyPrint n ++ " expects " ++ show arL
                        ++ " arguments but has been given " ++ show (length ans)
    Nothing -> fail $ "No such lock: " ++ prettyPrint n
  aids <- mapM getActor ans
  return $ TcLock n aids
evalLock (LockVar i) = return $ TcLockVar i
  
getActor :: ActorName -> TcBase ActorId
getActor (ActorName n) = do
  tm <- getTypeMap
  case lookupNamed actors n tm of
    Just aid -> return aid
    Nothing -> fail $ "getActor: No such actor: " ++ prettyPrint n
getActor (ActorTypeVar i) = return $ ActorTPVar i

-- From source

fromSrcType :: TypeMap -> Type -> TcType
fromSrcType tm (PrimType pt) = TcPrimT pt
fromSrcType tm (RefType rt) = TcRefT $ fromSrcRefType tm rt

fromSrcRefType :: TypeMap -> RefType -> TcRefType
fromSrcRefType tm (ClassRefType ct) = TcClsRefT $ fromSrcClsType tm ct
fromSrcRefType tm (ArrayType t mp) = 
   TcArrayT (fromSrcType tm t) (fmap (fromSrcPolicy tm) mp)
fromSrcRefType tm (TypeVariable i) = TcTypeVar i

fromSrcClsType :: TypeMap -> ClassType -> TcClassType
fromSrcClsType tm (ClassType iArgs) = TcClassT $ map (\(i,tas) -> (i, map (fromSrcTypeArg tm) tas)) iArgs

-- Ignore wildcards for now
fromSrcTypeArg :: TypeMap -> TypeArgument -> TcTypeArg
fromSrcTypeArg tm (ActualArg a) = fromSrcArgument tm a
fromSrcTypeArg tm _ = error "Wildcards not yet supported"

fromSrcArgument :: TypeMap -> NonWildTypeArgument -> TcTypeArg
fromSrcArgument tm (ActualType rt) = TcActualType $ fromSrcRefType tm rt
fromSrcArgument tm (ActualPolicy p) = TcActualPolicy $ fromSrcPolicy tm p
fromSrcArgument tm (ActualActor n)  = TcActualActor $ fromSrcActorN tm n
fromSrcArgument tm (ActualLockState ls) = TcActualLockState $ map (fromSrcLock tm) ls

fromSrcLock :: TypeMap -> Lock -> TcLock
fromSrcLock tm (LockVar i) = TcLockVar i
fromSrcLock tm (Lock n ans) = let aids = map (fromSrcActorName tm) ans in TcLock n aids

fromSrcActor :: TypeMap -> Actor -> TcActor
fromSrcActor tm (Actor an) = TcActor $ fromSrcActorName tm an
fromSrcActor tm (Var i) = TcVar i

fromSrcActorName :: TypeMap -> ActorName -> ActorId
fromSrcActorName tm (ActorName n) = fromSrcActorN tm n
fromSrcActorName tm (ActorTypeVar i) = ActorTPVar i

fromSrcActorN :: TypeMap -> Name -> ActorId
fromSrcActorN tm n = 
    case lookupNamed actors n tm of
     Just aid -> aid
     Nothing -> error "fromSrcActor: No such actor"

fromSrcPolicy :: TypeMap -> Policy -> TcPolicy
fromSrcPolicy tm (ExpName n) =
    case lookupNamed policies n tm of
      Just p -> p
      Nothing -> error "fromSrcPolicy: No such policy"
fromSrcPolicy tm (PolicyExp e) = fromSrcPolicyExp tm e

fromSrcPolicyExp :: TypeMap -> PolicyExp -> TcPolicy
fromSrcPolicyExp tm (PolicyLit cs) = TcPolicy $ map (fromSrcClause tm) cs
fromSrcPolicyExp tm (PolicyOf i) = TcRigidVar i
fromSrcPolicyExp tm (PolicyTypeVar i) = TcRigidVar i

fromSrcClause :: TypeMap -> Clause Actor -> TcClause TcActor
fromSrcClause tm (Clause h b) = TcClause (fromSrcActor tm h) (map (fromSrcAtom tm) b)

fromSrcAtom :: TypeMap -> Atom -> TcAtom
fromSrcAtom tm (Atom n as) = TcAtom n $ map (fromSrcActor tm) as


-------------------------------------------------------------------------------------      
getReadPolicy, getWritePolicy, getLockPolicy :: [Modifier] -> TcBase TcPolicy
getReadPolicy mods = 
    case [pol |Reads pol <- mods ] of -- !!0 -- Read Policy? what if no read policy?
      [pol] -> evalPolicy pol
      [] -> return bottom

getWritePolicy mods =
    case [pol | Writes pol <- mods] of
      [pol] -> evalPolicy pol
      [] -> return top

getLockPolicy mods =
    case [pol | Reads pol <- mods] of
      [pol] -> evalPolicy pol
      [] -> return top

getParamPolicy :: Ident -> [Modifier] -> TcBase TcPolicy
getParamPolicy i mods =
    case [pol | Reads pol <- mods ] of
       [pol] -> evalPolicy pol
       [] -> return $ ofPol i

getReturnPolicy :: [Modifier] -> [TcPolicy] -> TcBase TcPolicy
getReturnPolicy mods pPols = 
    case [pol | Reads pol <- mods ] of
      [pol] -> evalPolicy pol
      [] -> return $ foldl join bottom pPols

ofPol :: Ident -> TcPolicy
ofPol = TcRigidVar















{- Graveyard

lookupVar :: Name -> Tc VTypeInfo
lookupVar n = do
  tryVar n
    `orElse` tryField n
    `orElse` tryQual n
    `orElse` fail ("Unknown variable: " ++ prettyPrint n)

 where
   tryVar = tryVarField (vars <$> getEnv)
   tryField = tryVarField (fields <$> liftBase getTypeMap)

   --tryVarField :: Tc (Map Ident VTypeInfo) -> Name -> Tc (Maybe VTypeInfo)
   tryVarField recfetch (Name [i]) = do
     varMap <- recfetch
     case Map.lookup i varMap of
       Just vti  -> 
            if varType vti == actorT
            then -- We should check *which* actor
              do aid <- getActorId n
                 return $ Just $ vti { varType = actorIdT aid }
            else if varType vti == policyT
                 then do -- ... which policy
                   pol <- getPolicy n
                   return $ Just $ vti { varType = policyPolT pol }
                 else return $ Just vti
       Nothing -> return Nothing
   tryVarField _ n = return Nothing

   tryQual n = do
     tm <- liftBase getTypeMap
     return $ lookupNamed fields n tm
-}
